
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"

#include "errors.h"
#include "protocol.h"
#include "map.h"
#include "packages.h"
#include "players.h"
#include "logic.h"
#include "distancelist.h"
#include "bases.h"



/*  Main Map - Accessed through specific map_ commands*/
/*  Distance Map List */
/*  Player List */
/*  Base List */
/*  Package List - Accessed through  */
/*  Statistics (starting values mainly) */
/*  My Vitals */


global_info ginfo;


int logic_parsecommands(ll_rcvlist *cmd)
{
	int rval;
	_mapinfo_uint *tempmap;

	/* Decrement map for our current position */
	maplist_RemoveMap(ginfo.my_x,ginfo.my_y);


	while (cmd!=NULL)
	{
		/* Handling for self commands */
		if (cmd->playernum==ginfo.myuid)
		{
			/* Operation related to self */
			switch (cmd->command)
			{
			case CMD_RCV_MOVE:
				switch (cmd->op1)
				{
					case DIRECTION_N:
						(ginfo.my_y)++;
						break;
						
					case DIRECTION_S:
						(ginfo.my_y)--;
						break;
					case DIRECTION_E:
						(ginfo.my_x)++;
						break;
					case DIRECTION_W:
						(ginfo.my_x)--;
						break;
				}
				base_markvisited(ginfo.my_x,ginfo.my_y);
				break;
			case CMD_RCV_POS:
				ginfo.my_x=cmd->op1;
				ginfo.my_y=cmd->op2;
				base_markvisited(ginfo.my_x,ginfo.my_y);
				break;
			case CMD_RCV_PICK:
				ginfo.my_weightnow+=packages_total_getweight(cmd->op1);
				packages_self_pickedup(cmd->op1);
				break;
			case CMD_RCV_DROP:
				ginfo.my_weightnow-=packages_total_getweight(cmd->op1);
				packages_self_dropped(cmd->op1,ginfo.my_x,ginfo.my_y);
				break;
			}
		} else {
			/* Handling for only commands NOT on self */



		}

		/* Handling for all commands self or other */

		switch (cmd->command)
		{
		case CMD_RCV_MOVE:
			player_moved(cmd->playernum,cmd->op1);
			break;
		case CMD_RCV_POS:
			player_position(cmd->playernum,cmd->op1,cmd->op2);
			break;
		case CMD_RCV_PICK:
			packages_total_pickedup(cmd->op1,cmd->playernum);
			break;
		case CMD_RCV_DROP:
			{
				int px,py;

				if (player_getposition(cmd->playernum,&px,&py))
					packages_total_dropped(cmd->op1,px,py);
				else
					DEBUG_PRINT("Warning! Unknown or invalid player dropped a package!\n");
			}
			break;
		}


		
		/* Next command */
		cmd=cmd->next;
	}

#ifdef DEBUG_LOGIC
	DEBUG_PRINT("Creating a map for this position..\n");
#endif
	/* Create a map for our new position */
	maplist_ForcePermMap(&tempmap,ginfo.my_x,ginfo.my_y);

	return 0;

}


int logic_decision(sendcommand *execute)
{
	int rval;
	int bidmodifier;

/*
If you are in a deadly situation, bid high to avoid death.
Always avoid moving into a potentially "deadly" situation.
Avoid moving into a pushable situation if the cost of avoiding the push (beating the opponents bid) is not known to be less than an alternative path.  If there is no alternative path, outbid.
If someone moves into a deadly situation, push them with a known high bid, or reasonable guess.
If you have packages for the current square, drop them.
If you are on a package list with deliverable packages, pick up the most favorable combination you have room for.
If you have packages deliver the closest package first.
Head for the closest known list of packages.
Head for the base square that was visited least recently and is closest.

*/

#ifdef DEBUG_LOGIC
	DEBUG_PRINT("Into decision path..\n");
#endif

	bidmodifier=1;

	/* If you are in a deadly situation, bid high to avoid death. */
	if (map_IsDangerous(ginfo.my_x,ginfo.my_y))
	{
		if ((map_IsWater(ginfo.my_x-1,ginfo.my_y) && player_at(ginfo.my_x+1,ginfo.my_y)) ||
			(map_IsWater(ginfo.my_x+1,ginfo.my_y) && player_at(ginfo.my_x-1,ginfo.my_y)) ||
			(map_IsWater(ginfo.my_x,ginfo.my_y-1) && player_at(ginfo.my_x,ginfo.my_y+1)) ||
			(map_IsWater(ginfo.my_x,ginfo.my_y+1) && player_at(ginfo.my_x,ginfo.my_y-1)))
		{
			bidmodifier=50;
			/* More than prudent */
			if (bidmodifier>(ginfo.my_bidleft-10))
				bidmodifier=ginfo.my_bidleft-10;
			
			/* Oh well, even if we get out alive we dont have the moves to get points */
			if (bidmodifier<1)
				bidmodifier=1;
		}
	}


	/* Always avoid moving into a potentially "deadly" situation. */

	/*
	 *  Don't really have the time to impliment look ahead.  Will have to rely on the weighted distance map.
	 *
	 *
	 */


	/* Avoid moving into a pushable situation if the cost of avoiding the push (beating the opponents bid) is not known to be less than an alternative path.  If there is no alternative path, outbid. */

	/*
	 * No time, no time.  Weighted distance map again.
	 *
	 *
	 */

	/* If someone moves into a deadly situation, push them with a known high bid, or reasonable guess. */

	/*
	 *  Our default "push" will be 2.  We don't really care if we push them in or not all that much, since there could be
	 *   a LOT of other players not even battling (and thus saving moves).  If this is a deadly situation, the bid modifier overrides.
	 */ 
	{
		int direction;
		direction=DIRECTION_NONE;

		if (player_at(ginfo.my_x-1,ginfo.my_y) && map_IsWater(ginfo.my_x-2,ginfo.my_y))
			direction=DIRECTION_W;
		if (player_at(ginfo.my_x+1,ginfo.my_y) && map_IsWater(ginfo.my_x+2,ginfo.my_y))
			direction=DIRECTION_E;
		if (player_at(ginfo.my_x,ginfo.my_y-1) && map_IsWater(ginfo.my_x,ginfo.my_y-2))
			direction=DIRECTION_S;
		if (player_at(ginfo.my_x,ginfo.my_y+1) && map_IsWater(ginfo.my_x,ginfo.my_y+2))
			direction=DIRECTION_N;

		if (direction!=DIRECTION_NONE)
		{
			execute->bid=2;
			if (bidmodifier>execute->bid)
				execute->bid=bidmodifier;
			execute->command=CMD_SND_MOVE;
			execute->ops=(ll_oplist *)malloc(sizeof(ll_oplist));

			if (execute->ops==NULL)
				return ERR_OUTOFMEMORY;
			execute->ops->next=NULL;
			execute->ops->op=direction;

			return 0;
		}
	}



	/* If you have packages for the current square, drop them. */
	if (bidmodifier==1 && packages_self_havepackages())
	{
		ll_packages *best,*current;
		int distance;
		float factor,bestfactor;
		int direction,bestdirection;



		execute->bid=0;
		current=packages_self_getlist();
		while (current!=NULL)
		{
			if (current->destx==ginfo.my_x && current->desty==ginfo.my_y)
			{
				ll_oplist *newop;
				if (execute->bid==0)
				{
					execute->bid=1;
					execute->command=CMD_SND_DROP;
					execute->ops=NULL;
				}

				newop=(ll_oplist *)malloc(sizeof(ll_oplist));
				if (newop==NULL)
					return ERR_OUTOFMEMORY;

				newop->next=execute->ops;
				newop->op=current->packageid;
				execute->ops=newop;
			}
			current=current->next;
		}
		if (execute->bid>0)
		{
#ifdef DEBUG_LOGIC
			DEBUG_PRINT("If you have packages for the current square, drop them.\n");
#endif
			return 0;
		}
	}


	/* If you are on a package list with deliverable packages, pick up the most favorable combination. */
	if (bidmodifier==1) 
	{
		ll_packages *current;
		ll_distanceentry *current_dist;
		DISTENT distlist;
		int distance,direction;

		int extraweight=0;


		current=packages_total_getlist();
		distlist=distancelist_newlist();

		if (distlist==NULL)
			return ERR_OUTOFMEMORY;

		while (current!=NULL)
		{
			if (current->x==ginfo.my_x && current->y==ginfo.my_y && (current->weight+ginfo.my_weightnow <=ginfo.my_weightmax) && (current->holdingplayer!=ginfo.myuid))
			{
				distance=map_WeightedDistanceBetween(ginfo.my_x,ginfo.my_y,current->destx,current->desty);
				if (distance<1000001)
				{
					rval=distancelist_addentry(distlist,distance,(void *)current);
					if (FAILURE(rval))
						return rval;
				}

			}
			current=current->next;
		}

		
		execute->bid=0;
		current_dist=*distlist;

		while (current_dist!=NULL)
		{
			current=(ll_packages *)(current_dist->entry);
			if ((current->weight+ginfo.my_weightnow+extraweight <=ginfo.my_weightmax))
			{
				ll_oplist *newop;
				if (execute->bid==0)
				{
					execute->bid=1;
					execute->command=CMD_SND_PICK;
					execute->ops=NULL;
				}
				extraweight+=current->weight;

				newop=(ll_oplist *)malloc(sizeof(ll_oplist));
				if (newop==NULL)
					return ERR_OUTOFMEMORY;

				newop->next=execute->ops;
				newop->op=current->packageid;
				execute->ops=newop;
			}
			current_dist=current_dist->next;
		}
		
		/* Clear the list */
		distancelist_clearlist(distlist);

		if (execute->bid>0)
		{
#ifdef DEBUG_LOGIC
			DEBUG_PRINT("If you are on a package list with deliverable packages, pick up the most favorable combination.\n");
#endif
			return 0;
		}
	}


	/* If you have packages deliver the closest package first. */
	if (packages_self_havepackages())
	{
		ll_packages *best,*current;
		int distance;
		float factor,bestfactor;

		bestfactor=-1.0f;
		current=packages_self_getlist();
		execute->bid=0;

		while (current!=NULL)
		{
			distance=map_WeightedDistanceBetween(ginfo.my_x,ginfo.my_y,current->destx,current->desty);
			if (FAILURE(distance))
				return distance;

			if (distance==0)
				break;
			factor=(float)(current->weight)/(float)distance;
			if (factor>bestfactor)
			{
				best=current;
				bestfactor=factor;
				execute->bid=1;
			}
			current=current->next;
		}

		if (execute->bid!=0)
		{
#ifdef DEBUG_LOGIC
			DEBUG_PRINT("If you have packages deliver the closest package first.\n");
#endif

			execute->bid=bidmodifier;

			execute->command=CMD_SND_MOVE;
			execute->ops=(ll_oplist *)malloc(sizeof(ll_oplist));
			if (execute->ops==NULL)
				return ERR_OUTOFMEMORY;

			execute->ops->next=NULL;
			execute->ops->op=map_DirectionNextStepBetween(ginfo.my_x,ginfo.my_y,best->destx,best->desty);
			if (FAILURE(execute->ops->op))
				return execute->ops->op;

			return 0;
		}

	}




	
	/* Is the player/base ratio too high? - UNIMPLIMENTED */

	/* Head for the closest known list of packages. */
	{
		ll_baselist *current,*best;
		int distance,bestdistance;

		current=base_baselist();

		bestdistance=1000001;
		best=NULL;
		

		while (current!=NULL)
		{
			if (current->baseinfo.visited && packages_total_countdeliverable(current->baseinfo.x,current->baseinfo.y))
			{
				distance=map_WeightedDistanceBetween(ginfo.my_x,ginfo.my_y,current->baseinfo.x,current->baseinfo.y);
				if (FAILURE(distance))
					return distance;

				if (distance<bestdistance)
				{
					bestdistance=distance;
					best=current;
					
				}
			}
			current=current->next;
		}
		
		if (bestdistance<1000001)
		{
#ifdef DEBUG_LOGIC
			DEBUG_PRINT("Head for the closest known list of packages.\n");
#endif
			execute->bid=bidmodifier;
			execute->command=CMD_SND_MOVE;
			execute->ops=(ll_oplist *)malloc(sizeof(ll_oplist));
			if (execute->ops==NULL)
				return ERR_OUTOFMEMORY;
			execute->ops->next=NULL;
			execute->ops->op=map_DirectionNextStepBetween(ginfo.my_x,ginfo.my_y,best->baseinfo.x,best->baseinfo.y);
			if (FAILURE(execute->ops->op))
				return execute->ops->op;
			return 0;
		} else {
			/* If all bases have been visited in the past .. */
			current=base_baselist();

			while (current!=NULL)
			{
				if (!(current->baseinfo.visited))
					break;
				current=current->next;
			}
			if (current==NULL)
			{
				/* Perhaps someone is playing a moving game - start checking unknown package drops */
				ll_packages *pkg,*pkg_best;

				pkg=packages_total_getlist();

				bestdistance=1000001;
				pkg_best=NULL;
				
				while (pkg!=NULL)
				{
					if (pkg->holdingplayer==-1 && pkg->x!=-1 && pkg->y!=-1)
					{
						distance=map_WeightedDistanceBetween(ginfo.my_x,ginfo.my_y,pkg->x,pkg->y);
						if (FAILURE(distance))
							return distance;

						if (distance<bestdistance)
						{
							bestdistance=distance;
							pkg_best=pkg;
						}
					}
					pkg=pkg->next;
				}

				if (bestdistance<1000001)
				{
#ifdef DEBUG_LOGIC
					DEBUG_PRINT("Perhaps someone is playing a moving game - start checking unknown package drops.\n");
#endif
					execute->bid=bidmodifier;
					execute->command=CMD_SND_MOVE;
					execute->ops=(ll_oplist *)malloc(sizeof(ll_oplist));
					if (execute->ops==NULL)
						return ERR_OUTOFMEMORY;
					execute->ops->next=NULL;
					execute->ops->op=map_DirectionNextStepBetween(ginfo.my_x,ginfo.my_y,pkg_best->destx,pkg_best->desty);
					if (FAILURE(execute->ops->op))
						return execute->ops->op;
					return 0;
				}
					
			}
		}
	}


	/* Head for the base square that was visited least recently and is closest. */
	{
		ll_baselist *current,*best;
		int direction,bestdirection;
		int distance,bestdistance;
		int visited,bestvisited;

		current=base_baselist();

		bestdistance=1000001;
		bestvisited=ginfo.turn+1;
		best=NULL;

		while (current!=NULL)
		{
			if (current->baseinfo.visited<=bestvisited) 
			{
				distance=map_WeightedDistanceBetween(ginfo.my_x,ginfo.my_y,current->baseinfo.x,current->baseinfo.y);
				if (FAILURE(distance))
					return distance;
				if (current->baseinfo.visited<bestvisited || distance<bestdistance)
				{
					bestdistance=distance;
					best=current;
					bestvisited=current->baseinfo.visited;
				}

			}
			current=current->next;
		}
		
		if (bestdistance<1000001)
		{
#ifdef DEBUG_LOGIC
			DEBUG_PRINT("Head for the base square that was visited least recently and is closest.\n");
#endif
			execute->bid=bidmodifier;
			execute->command=CMD_SND_MOVE;
			execute->ops=(ll_oplist *)malloc(sizeof(ll_oplist));
			if (execute->ops==NULL)
				return ERR_OUTOFMEMORY;
			execute->ops->next=NULL;
			execute->ops->op=map_DirectionNextStepBetween(ginfo.my_x,ginfo.my_y,best->baseinfo.x,best->baseinfo.y);
			if (FAILURE(execute->ops->op))
				return execute->ops->op;
			return 0;
		}



	}

	/* When all else fails, idle? */
#ifdef DEBUG_LOGIC
	DEBUG_PRINT("Idling..\n");
#endif
	execute->bid=1;
	execute->command=CMD_SND_IDLE;
	execute->ops=NULL;

	return 0;		

}

