
#include <stdlib.h>
#include <string.h>

#include "debug.h"
#include "errors.h"


#include "protocol.h"
#include "players.h"
#include "logic.h"
#include "bases.h"
#include "map.h"

#define DANGER_WEIGHT 4
#define PLAYER_WEIGHT 1
#define AVOIDPUSH_WEIGHT 10


_mapinfo *CreateMap(int width,int height);
int AddMapRow(_mapinfo *map,unsigned int rownumber,unsigned char *rowdata);


int maplist_FindMap(_mapinfo_uint **mdata,int dx,int dy,int perm);
void maplist_RecursiveFill_Distance(_mapinfo_uint *map,int dx,int dy,int setdistance);
int map_GetDistance(_mapinfo_uint *dmap,int fx,int fy);
int maplist_PruneMaps(void);


void maplist_RecursiveBlock_Distance(_mapinfo_uint *map,int dx,int dy);
inline int maplist_GetDistance(_mapinfo_uint *map,int dx,int dy);
int maplist_SetWeighted_Distance(_mapinfo_uint *map,int dx,int dy,int setdistance);



_mapinfo *GameMap=NULL;
ll_maplist_uint *MapList=NULL;
int mapcount=0;


#define PRUNE_THRESH	50*1024*1024
#define HARD_THRESH		60*1024*1024


/*
 *  Map list manipulation functions
 *
 *
 */

int maplist_FindMap(_mapinfo_uint **mdata,int dx,int dy,int perm)
{
	ll_maplist_uint *current;

	current=MapList;

	while (current!=NULL)
	{
		if (current->x==dx && current->y==dy)
		{
			*mdata=current->map;
			(current->referencecount)++;
			if (current->lastturn<ginfo.turn)
			{
				if (perm)
					(current->lastturn)=2000000001;
				else
					(current->lastturn)=ginfo.turn;
			}
			return 0;
		}
		current=current->next;
	}
	*mdata=NULL;
	return 0;
}

int maplist_AddMap(_mapinfo_uint **mdata,int dx,int dy,int perm)
{
	int rval;
	ll_maplist_uint *newmap;

	/* Keep the size of the map list within reason */
	maplist_PruneMaps();

	newmap=(ll_maplist_uint *)malloc(sizeof(ll_maplist_uint));
	if (newmap==NULL)
		return ERR_OUTOFMEMORY;
	newmap->x=dx;
	newmap->y=dy;
	newmap->referencecount=1;
	if (perm)
		newmap->lastturn=2000000001;
	else
		newmap->lastturn=ginfo.turn;
	newmap->map=(_mapinfo_uint *)malloc(sizeof(_mapinfo_uint));
	if (newmap->map==NULL)
	{
		free(newmap);
		return ERR_OUTOFMEMORY;
	}

	newmap->map->width=GameMap->width;
	newmap->map->height=GameMap->height;
	newmap->map->data=(unsigned int *)malloc(sizeof(unsigned int)* (newmap->map->width) * (newmap->map->height));

	if (newmap->map->data==NULL)
	{
		free(newmap->map);
		free(newmap);
		return ERR_OUTOFMEMORY;
	}

	memset((void *)(newmap->map->data),0x7F,sizeof(unsigned int)* (newmap->map->width) * (newmap->map->height));

	/* Fill map with data */
	(newmap->map->data)[(dy-1)*(newmap->map->width)+(dx-1)]=0;
	maplist_RecursiveBlock_Distance(newmap->map,dx,dy);

	/* Add to list */

	if (MapList!=NULL)
		MapList->prev=newmap;

	newmap->next=MapList;
	newmap->prev=NULL;
	MapList=newmap;

	mapcount++;
	
	*mdata=newmap->map;
	return 0;
}


void maplist_RecursiveBlock_Distance(_mapinfo_uint *map,int dx,int dy)
{
	int directionmask,mydistance,mindistance,direction;
	int d_e,d_w,d_n,d_s;

	directionmask=0x00;
	mydistance=(map->data)[(dy-1)*(map->width)+(dx-1)];

	d_n=maplist_SetWeighted_Distance(map,dx,dy+1,mydistance+1);
	if (d_n==1000001)
		directionmask|=0x01;
	d_s=maplist_SetWeighted_Distance(map,dx,dy-1,mydistance+1);
	if (d_s==1000001)
		directionmask|=0x02;
	d_e=maplist_SetWeighted_Distance(map,dx+1,dy,mydistance+1);
	if (d_e==1000001)
		directionmask|=0x04;
	d_w=maplist_SetWeighted_Distance(map,dx-1,dy,mydistance+1);
	if (d_w==1000001)
		directionmask|=0x08;


	/* Mask N 1,S 2,E 4,W 8 */
	while (directionmask!=0x0F)
	{
		mindistance=1000001;
		direction=0x00;

		/* find the shortest next step */
		if (!(directionmask & 0x01))
		{
			if (d_n>=1000001)
				directionmask|=0x01;
			else
			{
				if (d_n<mindistance)
				{
					mindistance=d_n;
					direction=0x01;
				}
			}
		}
		if (!(directionmask & 0x02))
		{
			if (d_s>=1000001)
				directionmask|=0x02;
			else
			{
				if (d_s<mindistance)
				{
					mindistance=d_s;
					direction=0x02;
				}
			}
		}
		if (!(directionmask & 0x04))
		{
			if (d_e>=1000001)
				directionmask|=0x04;
			else
			{
				if (d_e<mindistance)
				{
					mindistance=d_e;
					direction=0x04;
				}
			}
		}
		if (!(directionmask & 0x08))
		{
			if (d_w>=1000001)
				directionmask|=0x08;
			else
			{
				if (d_w<mindistance)
				{
					mindistance=d_w;
					direction=0x08;
				}
			}
		}

		switch (direction)
		{
		case 0x01:
			maplist_RecursiveBlock_Distance(map,dx,dy+1);
			directionmask|=0x01;
			break;
		case 0x02:
			maplist_RecursiveBlock_Distance(map,dx,dy-1);
			directionmask|=0x02;
			break;
		case 0x04:
			maplist_RecursiveBlock_Distance(map,dx+1,dy);
			directionmask|=0x04;
			break;
		case 0x08:
			maplist_RecursiveBlock_Distance(map,dx-1,dy);
			directionmask|=0x08;
			break;
		}

	}
}

inline int maplist_GetDistance(_mapinfo_uint *map,int dx,int dy)
{
	int maptile;

	if (dx<1 || dx>map->width)
		return 1000001;
	if (dy<1 || dy>map->height)
		return 1000001;
	
	maptile=map_GetTile(dx,dy);
		
	if (maptile==MAPTILE_WALL || maptile==MAPTILE_WATER)
		return 1000001;

	return (map->data)[(dy-1)*(map->width)+(dx-1)];
}




int maplist_SetWeighted_Distance(_mapinfo_uint *map,int dx,int dy,int setdistance)
{
	unsigned char maptile;
	int mt_n,mt_e,mt_w,mt_s;
	int pnum;

	if (dx<1 || dx>map->width)
		return 1000001;
	if (dy<1 || dy>map->height)
		return 1000001;

	maptile=map_GetTile(dx,dy);
		
	if (maptile==MAPTILE_WALL || maptile==MAPTILE_WATER)
		return 1000001;

	mt_n=map_GetTile(dx,dy+1);
	mt_s=map_GetTile(dx,dy-1);
	mt_e=map_GetTile(dx+1,dy);
	mt_w=map_GetTile(dx-1,dy);


	/* Weighted distance */
	if (mt_w==MAPTILE_WATER && mt_e==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;
	if (mt_e==MAPTILE_WATER && mt_w==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;
	if (mt_s==MAPTILE_WATER && mt_n==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;
	if (mt_n==MAPTILE_WATER && mt_s==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;

	pnum=player_at(dx-1,dy);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;
	pnum=player_at(dx+1,dy);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;
	pnum=player_at(dx,dy-1);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;
	pnum=player_at(dx,dy+1);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;

	pnum=player_at(dx,dy);
	if (pnum && (pnum!=ginfo.myuid))
	{
		if (ginfo.my_x==dx-1 && ginfo.my_y==dy)
			setdistance+=AVOIDPUSH_WEIGHT;
		if (ginfo.my_x==dx+1 && ginfo.my_y==dy)
			setdistance+=AVOIDPUSH_WEIGHT;
		if (ginfo.my_x==dx && ginfo.my_y==dy-1)
			setdistance+=AVOIDPUSH_WEIGHT;
		if (ginfo.my_x==dx && ginfo.my_y==dy+1)
			setdistance+=AVOIDPUSH_WEIGHT;
	}




	if ((map->data)[(dy-1)*(map->width)+(dx-1)]<=setdistance)
		return 1000001;


	(map->data)[(dy-1)*(map->width)+(dx-1)]=setdistance;

	return (map->data)[(dy-1)*(map->width)+(dx-1)];
}

void maplist_RecursiveFill_Distance(_mapinfo_uint *map,int dx,int dy,int setdistance)
{
	unsigned char maptile;
	int mt_n,mt_e,mt_w,mt_s;
	int pnum;

	if (dx<1 || dx>map->width)
		return;
	if (dy<1 || dy>map->height)
		return;


	maptile=map_GetTile(dx,dy);
		
	if (maptile==MAPTILE_WALL || maptile==MAPTILE_WATER)
		return;

	mt_n=map_GetTile(dx,dy+1);
	mt_s=map_GetTile(dx,dy-1);
	mt_e=map_GetTile(dx+1,dy);
	mt_w=map_GetTile(dx-1,dy);


	/* Weighted distance */
	if (mt_w==MAPTILE_WATER && mt_e==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;
	if (mt_e==MAPTILE_WATER && mt_w==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;
	if (mt_s==MAPTILE_WATER && mt_n==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;
	if (mt_n==MAPTILE_WATER && mt_s==MAPTILE_EMPTY)
		setdistance+=DANGER_WEIGHT;

	pnum=player_at(dx-1,dy);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;
	pnum=player_at(dx+1,dy);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;
	pnum=player_at(dx,dy-1);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;
	pnum=player_at(dx,dy+1);
	if (pnum && (pnum!=ginfo.myuid))
		setdistance+=PLAYER_WEIGHT;

	pnum=player_at(dx,dy);
	if (pnum && (pnum!=ginfo.myuid))
	{
		if (ginfo.my_x==dx-1 && ginfo.my_y==dy)
			setdistance+=AVOIDPUSH_WEIGHT;
		if (ginfo.my_x==dx+1 && ginfo.my_y==dy)
			setdistance+=AVOIDPUSH_WEIGHT;
		if (ginfo.my_x==dx && ginfo.my_y==dy-1)
			setdistance+=AVOIDPUSH_WEIGHT;
		if (ginfo.my_x==dx && ginfo.my_y==dy+1)
			setdistance+=AVOIDPUSH_WEIGHT;
	}




	if ((map->data)[(dy-1)*(map->width)+(dx-1)]<=setdistance)
		return;


	(map->data)[(dy-1)*(map->width)+(dx-1)]=setdistance;

	/* N */
	maplist_RecursiveFill_Distance(map,dx,dy+1,setdistance+1);
	/* S */
	maplist_RecursiveFill_Distance(map,dx,dy-1,setdistance+1);
	/* E */
	maplist_RecursiveFill_Distance(map,dx+1,dy,setdistance+1);
	/* W */
	maplist_RecursiveFill_Distance(map,dx-1,dy,setdistance+1);

}


int maplist_ForceMap(_mapinfo_uint **mdata,int dx,int dy)
{
	int rval;

	*mdata=NULL;

	rval=maplist_FindMap(mdata,dx,dy,0);
	if (FAILURE(rval))
		return rval;

	if (*mdata!=NULL)
		return 0;

	return maplist_AddMap(mdata,dx,dy,0);
}

int maplist_ForcePermMap(_mapinfo_uint **mdata,int dx,int dy)
{
	int rval;

	*mdata=NULL;

	rval=maplist_FindMap(mdata,dx,dy,1);
	if (FAILURE(rval))
		return rval;

	if (*mdata!=NULL)
		return 0;

	return maplist_AddMap(mdata,dx,dy,1);
}


int maplist_PruneMaps(void)
{
	int usage;
	ll_maplist_uint *current,*worst;
	int worstturn;


	
	usage=mapcount*(GameMap->width)*(GameMap->height)*sizeof(unsigned int);

	while (usage>PRUNE_THRESH)
	{
		current=MapList;
		worst=NULL;
		worstturn=2000000000;

		while (current!=NULL)
		{
			if (current->lastturn<worstturn)
			{
				worstturn=current->lastturn;
				worst=current;
			}
			current=current->next;
		}
		if (worst==NULL)
			return 0; /* all maps unremovable */
		if (worst->next!=NULL)
			worst->next->prev=worst->prev;

		if (worst->prev!=NULL)
			worst->prev->next=worst->next;
		else
			MapList=worst->next;

		free(worst->map->data);
		free(worst->map);
		free(worst);

		mapcount--;

		usage=mapcount*(GameMap->width)*(GameMap->height)*sizeof(unsigned int);
	}

	return 0;
}

int maplist_RemoveMap(int dx,int dy)
{
	ll_maplist_uint *current;

	current=MapList;

	while (current!=NULL)
	{
		if (current->x==dx && current->y==dy)
			break;
		current=current->next;
	}
	if (current==NULL)
		return 0; /* Not found */


	(current->referencecount)--;
	if (current->referencecount>0)
		return 0; /* Still a reference to this */

	mapcount--;

	free(current->map->data);
	free(current->map);

	if (current->next!=NULL)
		current->next->prev=current->prev;
	if (current->prev!=NULL)
		current->prev->next=current->next;
	else
		MapList=current->next;

	free(current);


	return 0;

}


int map_IsDangerous(int x,int y)
{
	if (map_GetTile(x-1,y)==MAPTILE_WATER && map_GetTile(x+1,y)==MAPTILE_EMPTY)
		return 1;
	if (map_GetTile(x+1,y)==MAPTILE_WATER && map_GetTile(x-1,y)==MAPTILE_EMPTY)
		return 1;
	if (map_GetTile(x,y-1)==MAPTILE_WATER && map_GetTile(x,y+1)==MAPTILE_EMPTY)
		return 1;
	if (map_GetTile(x,y+1)==MAPTILE_WATER && map_GetTile(x,y-1)==MAPTILE_EMPTY)
		return 1;
	return 0;
}


inline int map_GetTile(int x,int y)
{
	if (x<1 || x>GameMap->width)
		return MAPTILE_WALL;
	if (y<1 || y>GameMap->height)
		return MAPTILE_WALL;

	switch((GameMap->data)[(y-1)*(GameMap->width)+(x-1)])
	{
	case '~':
		return MAPTILE_WATER;
		break;
	case '.':
	case '@':
		return MAPTILE_EMPTY;
		break;
	default:
		return MAPTILE_WALL;
		break;
	}
	return MAPTILE_WALL;
}




int map_IsWater(int x,int y)
{
	if (x<1 || x>GameMap->width)
		return 0;
	if (y<1 || y>GameMap->height)
		return 0;

	if ((GameMap->data)[(y-1)*(GameMap->width)+(x-1)]=='~')
		return 1;
	return 0;
}


int map_DirectionNextStepBetween(int fx,int fy,int tx,int ty)
{
	int rval;
	_mapinfo_uint *dmap;
	int testdistance,shortdistance;
	int direction;
	int cx,cy;

	dmap=NULL;
	/* Make a map with the same tx,ty */
	rval=maplist_ForceMap(&dmap,tx,ty);
	if (FAILURE(rval))
		return rval;
	if (dmap!=NULL)
	{
		/* Found */
		direction=DIRECTION_NONE;
		shortdistance=1000001;

		/* N */
		testdistance=map_GetDistance(dmap,fx,fy+1);
		if (testdistance<shortdistance)
		{
			shortdistance=testdistance;
			direction=DIRECTION_N;
		}
		/* S */
		testdistance=map_GetDistance(dmap,fx,fy-1);
		if (testdistance<shortdistance)
		{
			shortdistance=testdistance;
			direction=DIRECTION_S;
		}
		/* E */
		testdistance=map_GetDistance(dmap,fx+1,fy);
		if (testdistance<shortdistance)
		{
			shortdistance=testdistance;
			direction=DIRECTION_E;
		}
		/* W */
		testdistance=map_GetDistance(dmap,fx-1,fy);
		if (testdistance<shortdistance)
		{
			shortdistance=testdistance;
			direction=DIRECTION_W;
		}

		/* Let pruning take care of this */
		/*
		maplist_RemoveMap(tx,ty);
		*/

		if (direction!=DIRECTION_NONE)
			return direction;
	}
	
	return ERR_UNREACHABLE;


}

/*
 * Values > 1000000 are unreachable
 *
 *
 *
 */


int map_WeightedDistanceBetween(int fx,int fy,int tx,int ty)
{
	int rval;
	_mapinfo_uint *dmap;
	int testdistance,shortdistance=1000001;

	/* Search for a map with the same fx,fy */

	dmap=NULL;
	rval=maplist_ForceMap(&dmap,fx,fy);
	if (FAILURE(rval))
		return rval;

	rval=(dmap->data)[(ty-1)*(dmap->width)+(tx-1)];

#ifdef DEBUG_MAP
	map_PrintDistanceMap(dmap);
#endif

	maplist_RemoveMap(fx,fy);

	return rval;
}

int map_GetDistance(_mapinfo_uint *dmap,int fx,int fy)
{
	if (fx<1 || fx>dmap->width)
		return 1000001;
	if (fy<1 || fy>dmap->height)
		return 1000001;

	return (dmap->data)[(fy-1)*(dmap->width)+(fx-1)];

}


int map_InitializeMap(mapdata *md)
{
	int rval;
	ll_rowdata *current_row;
	ll_baselist *current_base;
	int basecount;

	GameMap=CreateMap(md->width,md->height);

	if (GameMap==NULL)
		return ERR_OUTOFMEMORY;


	for (current_row=md->rows;current_row!=NULL;current_row=current_row->next)
	{
		rval=AddMapRow(GameMap,current_row->rownumber,current_row->rowdata);

		if (FAILURE(rval))
		{
			DEBUG_PRINT("Error calling AddMapRow().\n");
			return rval;
		}
	}


	/* Verify memory conservation */
/*
	current_base=base_baselist();
	while (current_base!=NULL)
	{
		basecount++;
		current_base=current_base->next;
	}
*/

	/*

	if ((basecount*sizeof(unsigned int)*(GameMap->width)*(GameMap->height)) < (HARD_THRESH))
	{
		_mapinfo_uint *mdata;
		current_base=base_baselist();
		while (current_base!=NULL)
		{
#ifdef DEBUG_INIT
			DEBUG_PRINT("Creating base map..\n");
#endif
*/
	
			/* For quicker lookup, bases can be forced to have a perm distance map */

	/*
			maplist_ForcePermMap(&mdata,current_base->baseinfo.x,current_base->baseinfo.y);
			current_base=current_base->next;
		}
	}
	*/



	return 0;
}




_mapinfo *CreateMap(int width,int height)
{
	_mapinfo *map;

	map=(_mapinfo *)malloc(sizeof(_mapinfo));

	if (map==NULL)
		return NULL;


	map->width=width;
	map->height=height;

	map->data=(unsigned char *)malloc(width*height*sizeof(unsigned char));

	if (map->data==NULL)
	{
		free(map);
		return NULL;
	}

	memset(map->data,0,width*height*sizeof(unsigned char));

	return map;
}


int AddMapRow(_mapinfo *map,unsigned int rownumber,unsigned char *rowdata)
{
	unsigned char *rowptr;
	int bytecnt;


	if (rownumber>map->height)
		return ERR_INDEXOUTOFBOUNDS;


	rowptr=(map->data)+((map->width)*(rownumber-1));

	memcpy(rowptr,rowdata,sizeof(unsigned char)*(map->width));

	for (bytecnt=0;bytecnt<(map->width);bytecnt++,rowptr++)
		if (*rowptr=='@')
		{
			/* Add base to the list */
			base_addbase(bytecnt+1,rownumber);
		}
	

	return 0;
}




int map_PrintMap(mapdata *map)
{
	int buffersize=map->width+2;
	char *linebuffer;
	ll_rowdata *walk;
	int line;
	

	if (buffersize<1024)
		buffersize=1024;

	linebuffer=(char *)malloc(buffersize);

	if (linebuffer==NULL)
		return ERR_OUTOFMEMORY;

	sprintf(linebuffer,"Width: %d  Height: %d\n",map->width,map->height);
	DEBUG_PRINT(linebuffer);

	for (line=map->height;line>0;line--)
	{
		walk=map->rows;
		while (walk!=NULL)
		{
			if (walk->rownumber==line)
			{
				sprintf(linebuffer,"%s\n",walk->rowdata);
				DEBUG_PRINT(linebuffer);
				break;
			}
			walk=walk->next;
		}
		if (walk==NULL)
		{
			sprintf(linebuffer,"ERROR!  Line %d not found!!\n",line);
			DEBUG_PRINT(linebuffer);
			return ERR_DATA_NONSENSE;
		}
	}


	return 0;
}

int map_PrintDistanceMap(_mapinfo_uint *dmap)
{
	int x,y;
	unsigned int distance;
	static char obuf[32];

	for (y=dmap->height;y>0;y--)
	{
		for (x=0;x<dmap->width;x++)
		{
			distance=(dmap->data)[((y-1)*dmap->width+(x-1))];
			if (distance>1000000)
				DEBUG_PRINT("#");
			else
			{
				sprintf(obuf,"%d",distance % 10);
				DEBUG_PRINT(obuf);
			}
		}
		DEBUG_PRINT("\n");
	}
	DEBUG_PRINT("\n\n");

	return 0;
}


int map_PrintAnsiMap(void)
{
	int x,y;
	static char obuf[32];
	ll_player *current;

	/* Move to home position */
	sprintf(obuf,"\e[0;0H"); 
	DEBUG_PRINT(obuf);


	for (y=GameMap->height;y>0;y--)
	{
		for (x=1;x<=GameMap->width;x++)
		{
			/* If a player is at this position, show it */
			current=player_getlist();
			while (current!=NULL)
			{
				if (current->data.x==x && current->data.y==y && current->data.lastturnseen==ginfo.turn)
				{
					sprintf(obuf,"%d",current->data.playerid %10);
					DEBUG_PRINT(obuf);
					break;
				}
				current=current->next;
			}
			if (current==NULL)
			{
				obuf[0]=(GameMap->data)[(y-1)*GameMap->width+(x-1)];
				obuf[1]=0x00;
				DEBUG_PRINT(obuf);
			}
		}
		DEBUG_PRINT("\e[K\n");
	}
	sprintf(obuf,"Score:%d\n",ginfo.my_score);
	DEBUG_PRINT(obuf);

/*	sprintf(obuf,"\e[%dA",GameMap->height+1); */
	return 0;
}



