/*
 *  Network related functions
 *
 *
 *
 */ 

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "debug.h"

#include "network.h"
#include "errors.h"


int resolve_address(struct sockaddr * addr, char *hostname, unsigned short port);



int game_sock=-1;



int net_connect(char *host,int port)
{
	unsigned short real_port;
	struct protoent *tcp_ent;
	int proto_number;
	struct sockaddr server_addr;

	if (port<0 || port>65535)
		return ERR_NETWORK_PORTOUTOFRANGE;

	real_port=(unsigned short)(port & 0xFFFF);


	if (game_sock>0)
	{
		/* Kill the previous connection */
		shutdown(game_sock,2);
		game_sock=-1;
	}

	/* Get the protocol number for TCP */
	tcp_ent=getprotobyname("TCP");

	if (tcp_ent==NULL)
	{
		/* This system's protocols file is either out of date or mangled
		 *
		 *  Do we really care anyway?  TCP is 6.
		 */
		proto_number=6;
	} else {
		proto_number=tcp_ent->p_proto;
	}


	/* Create a TCP socket */
	game_sock=socket(PF_INET,SOCK_STREAM,proto_number);

	if (game_sock<0)
		return ERR_NETWORK_SOCKETCREATE;

	if (resolve_address(&server_addr,host,real_port))
	{
		shutdown(game_sock,2);
		game_sock=-1;
		return ERR_NETWORK_COULDNOTRESOLVE;
	}

	if (connect(game_sock,&server_addr,sizeof(server_addr)))
	{
		shutdown(game_sock,2);
		game_sock=-1;
		return ERR_NETWORK_CONNECT;
	}

	/* Non blocking */
	/*
	if (fcntl(game_sock,F_SETFL,O_NONBLOCK))
	{
		DEBUG_PRINT("Non blocking mode was NOT set.\n");
	}
	*/



	return 0;

}


/*
 *  Closes a network connection
 *
 */

int net_close(void)
{
	if (game_sock<0)
		return ERR_NETWORK_NOTCONNECTED;
	shutdown(game_sock,2);
	game_sock=-1;
	return 0;
}

/*
 *  Gets as much waiting data as will fit in the supplied buffer
 *
 */


int net_getblock(char *buffer, int blen)
{
	int rval;

	if (game_sock<0)
		return ERR_NETWORK_NOTCONNECTED;

	rval=recv(game_sock,buffer,blen,0);

	if (rval==-1)
	{
		if (errno==EWOULDBLOCK)
			return 0;
		return ERR_NETWORK_READERROR;
	}
	if (!rval)
		return ERR_NETWORK_NOTCONNECTED;
	return rval;
}

/*
 *  Sends as much data as possible.  May not send the entire block.
 *
 *
 */

int net_sendblock(char *buffer,int blen)
{
	int rval;

	if (game_sock<0)
		return ERR_NETWORK_NOTCONNECTED;
	
	rval=send(game_sock,buffer,blen,0);
	if (rval==-1)
	{
		if (errno==EWOULDBLOCK)
			return 0;
		return ERR_NETWORK_SENDERROR;
	}
	return rval;
}






int resolve_address(struct sockaddr * addr, char *hostname, unsigned short port) 
{
  struct sockaddr_in *address;
  struct hostent *host;

  address = (struct sockaddr_in *)addr;
  memset( (char *)address,0,sizeof(struct sockaddr_in) );
  address->sin_family = AF_INET;
  address->sin_port = htons(port);
  address->sin_addr.s_addr = inet_addr(hostname);
  if ( (int)address->sin_addr.s_addr == -1) 
  {
    host = gethostbyname(hostname);
    if (host) 
    {
      memcpy( (char *)&address->sin_addr, host->h_addr,
      host->h_length);
    }
     else 
     {
		return ERR_NETWORK_COULDNOTRESOLVE;
     }
  }
  return 0;
}



